"""
Print compilation information for the grav_sim C library.

Usage:
    python -m grav_sim

This script searches for the compiled C library in the parent directory of the current file,
loads it using ctypes, and calls the `print_compilation_info` function defined in the C library.
User can run this script to verify the installation of the grav_sim library.
"""

import ctypes
from pathlib import Path


def main() -> None:
    search_path = Path(__file__).parent.parent
    c_lib_files = [str(p) for p in search_path.rglob("*libgrav_sim*")]
    if len(c_lib_files) == 0:
        raise FileNotFoundError(f"C library not found from path: {search_path}")

    c_lib_path = c_lib_files[0]
    c_lib: ctypes.CDLL = ctypes.cdll.LoadLibrary(c_lib_path)
    c_lib.print_compilation_info()
    print(f"C library location: {c_lib_path}")


if __name__ == "__main__":
    main()



"""
Application user interface (API) for gravity simulator

Usage:
    from grav_sim import GravitySimulatorAPI
    gs = GravitySimulatorAPI()
"""

import ctypes
import warnings
from pathlib import Path
from typing import Optional, Tuple

import numpy as np

from . import plotting
from . import utils
from . import parameters
from .simulator import Simulator
from .system import System, CosmologicalSystem


class GravitySimulatorAPI:
    """Gravity simulator API"""

    def __init__(self, c_lib_path: Optional[str] = None) -> None:
        """
        Initialize gravity simulator API

        Parameters
        ----------
        c_lib_path : str, optional
            Path to C library, by default None
        """
        if c_lib_path is not None:
            self.c_lib = utils.load_c_lib(Path(c_lib_path))
        else:
            self.c_lib = utils.load_c_lib()
        utils.initialize_c_lib(self.c_lib)

        # System
        self.BUILT_IN_SYSTEMS = System.BUILT_IN_SYSTEMS

        # Plotting
        self.SOLAR_SYSTEM_COLORS = plotting.SOLAR_SYSTEM_COLORS
        self.plot_quantity_against_time = plotting.plot_quantity_against_time
        self.plot_2d_trajectory = plotting.plot_2d_trajectory
        self.plot_3d_trajectory = plotting.plot_3d_trajectory

        # Simulator
        self.simulator = Simulator(c_lib=self.c_lib)
        self.DAYS_PER_YEAR = self.simulator.DAYS_PER_YEAR
        self.launch_simulation = self.simulator.launch_simulation
        self.launch_cosmological_simulation = (
            self.simulator.launch_cosmological_simulation
        )

        # Parameters
        self.AVAILABLE_ACCELERATION_METHODS = (
            parameters.AccelerationParam.AVAILABLE_ACCELERATION_METHODS
        )
        self.AVAILABLE_INTEGRATORS = parameters.IntegratorParam.AVAILABLE_INTEGRATORS
        self.FIXED_STEP_SIZE_INTEGRATORS = (
            parameters.IntegratorParam.FIXED_STEP_SIZE_INTEGRATORS
        )
        self.ADAPTIVE_STEP_SIZE_INTEGRATORS = (
            parameters.IntegratorParam.ADAPTIVE_STEP_SIZE_INTEGRATORS
        )
        self.AVAILABLE_OUTPUT_METHODS = parameters.OutputParam.AVAILABLE_OUTPUT_METHODS
        self.AVAILABLE_OUTPUT_DTYPE = parameters.OutputParam.AVAILABLE_OUTPUT_DTYPE

    def get_new_system(self) -> System:
        """Create a gravitational system

        Returns
        -------
        System object
        """
        return System(c_lib=self.c_lib)

    def get_new_cosmological_system(self) -> CosmologicalSystem:
        """Create a cosmological system

        Returns
        -------
        CosmologicalSystem object
        """
        return CosmologicalSystem(c_lib=self.c_lib)

    def load_system(
        self,
        file_path: str | Path,
    ) -> System:
        """Load system from a CSV file

        Parameters
        ----------
        file_path : str
            File path to load the system from
        """
        return System.load_system(self.c_lib, file_path)

    def get_built_in_system(self, system_name: str) -> System:
        """Get a built-in gravitational system

        Parameters
        ----------
        system_name : str
            Name of the built-in system to be loaded.
        """
        return System.get_built_in_system(self.c_lib, system_name)

    @staticmethod
    def get_new_parameters() -> Tuple[
        parameters.AccelerationParam,
        parameters.IntegratorParam,
        parameters.OutputParam,
        parameters.Settings,
    ]:
        """Create new simulation parameters

        Returns
        -------
        Tuple of acceleration, integrator, output, and settings parameters
        """
        acceleration_param = parameters.AccelerationParam()
        integrator_param = parameters.IntegratorParam()
        output_param = parameters.OutputParam()
        settings = parameters.Settings()

        return acceleration_param, integrator_param, output_param, settings

    def days_to_years(self, days: float | np.ndarray) -> float | np.ndarray:
        return days / self.simulator.DAYS_PER_YEAR

    def years_to_days(self, years: float | np.ndarray) -> float | np.ndarray:
        return years * self.simulator.DAYS_PER_YEAR

    @staticmethod
    def read_csv_data(
        output_dir: str | Path,
    ) -> Tuple[float, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Read CSV snapshots from the output directory,
        assuming number of particles and particle_ids
        stays the same

        Parameters
        ----------
        output_dir : str | Path
            Output directory path

        Returns
        -------
        G : float
            Gravitational constant
        time : np.ndarray
            Simulation time of each snapshot
        dt : np.ndarray
            Time step of each snapshot
        particle_ids : np.ndarray
            1D array of Particle IDs
        sol_state : np.ndarray
            3D array of solution state for each snapshot, with shape
            (num_snapshots, num_particles, 7) being
            [m, x, y, z, vx, vy, vz]
        """
        output_dir = Path(output_dir)
        if not output_dir.is_dir():
            raise FileNotFoundError(f"Output directory not found: {output_dir}")

        snapshot_files = sorted(output_dir.glob("snapshot_*.csv"))
        if len(snapshot_files) == 0:
            raise FileNotFoundError(f"No snapshot files found in: {output_dir}")

        G = -1.0
        time = np.zeros(len(snapshot_files), dtype=np.float64)
        dt = np.zeros(len(snapshot_files), dtype=np.float64)

        for i, snapshot_file in enumerate(snapshot_files):
            # Read the metadata
            with open(snapshot_file, "r") as file:
                read_metadata_num_particles = False
                read_metadata_G = False
                read_metadata_time = False
                read_metadata_dt = False
                for line in file:
                    line = line.strip()

                    if line.startswith("#"):
                        if line.startswith("# num_particles"):
                            if i == 0:
                                num_particles = int(line.split(":")[1].strip())
                            elif num_particles != int(line.split(":")[1].strip()):
                                raise ValueError(
                                    f"Number of particles changed from {num_particles} to {int(line.split(':')[1].strip())}"
                                )
                            read_metadata_num_particles = True
                        elif line.startswith("# G"):
                            G = float(line.split(":")[1].strip())
                            read_metadata_G = True
                        elif line.startswith("# time"):
                            time[i] = float(line.split(":")[1].strip())
                            read_metadata_time = True
                        elif line.startswith("# dt"):
                            dt[i] = float(line.split(":")[1].strip())
                            read_metadata_dt = True

                    if (
                        read_metadata_num_particles
                        and read_metadata_G
                        and read_metadata_time
                        and read_metadata_dt
                    ):
                        break

        # Read the data
        particle_ids = np.zeros(num_particles, dtype=np.int32)
        sol_state = np.zeros((len(snapshot_files), num_particles, 7), dtype=np.float64)
        for i, snapshot_file in enumerate(snapshot_files):
            data = np.genfromtxt(snapshot_file, delimiter=",", skip_header=5)
            if i == 0:
                particle_ids = data[:, 0].astype(np.int32)
                particle_ids = np.sort(particle_ids)
                _, num_duplicates = np.unique(particle_ids, return_counts=True)
                if np.any(num_duplicates > 1):
                    raise ValueError(
                        f"Particle IDs are not unique. Particle IDs: {particle_ids}"
                    )

            snapshot_particle_ids = data[:, 0].astype(np.int32)

            # Sort the data by particle IDs
            sorted_indices = np.argsort(snapshot_particle_ids)
            data = data[sorted_indices]

            # Check if the particle IDs match
            if not np.array_equal(particle_ids, snapshot_particle_ids):
                raise ValueError(
                    f"Particle IDs do not match in snapshot {i + 1}: {snapshot_file}"
                )

            # Store the data
            sol_state[i, :, :] = data[:, 1:]

        return G, time, dt, particle_ids, sol_state

    @staticmethod
    def delete_snapshots(
        output_dir: str | Path,
    ):
        """Delete all snapshots in the output directory

        Parameters
        ----------
        output_dir : str | Path
            Output directory path
        """
        output_dir = Path(output_dir)
        if not output_dir.is_dir():
            raise FileNotFoundError(f"Output directory not found: {output_dir}")

        snapshot_files = sorted(output_dir.glob("snapshot_*.csv"))
        for snapshot_file in snapshot_files:
            snapshot_file.unlink()

    def compute_energy(
        self, sol_state: np.ndarray, G: float
    ) -> np.ndarray:
        """Compute the total energy of the system

        Parameters
        ----------
        sol_state : np.ndarray
            3D array of solution state for each snapshot, with shape
            (num_snapshots, num_particles, 7) being
            [m, x, y, z, vx, vy, vz]
        G : float
            Gravitational constant

        Returns
        -------
        energy : np.ndarray
            1D array of total energy for each snapshot
        """
        # Check the dimension and shape of sol_state
        if len(sol_state.shape) != 3:
            raise ValueError("sol_state must be a 3D array")

        if sol_state.shape[2] != 7:
            raise ValueError(
                "sol_state must have shape (num_snapshots, num_particles, 7)"
            )

        # Compute the total energy
        energy = np.zeros(sol_state.shape[0], dtype=np.float64)
        self.c_lib.compute_energy_python(
            energy.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
            ctypes.c_double(G),
            sol_state.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
            ctypes.c_int32(sol_state.shape[0]),
            ctypes.c_int32(sol_state.shape[1]),
        )

        return energy

    def compute_linear_momentum(
        self,
        sol_state: np.ndarray,
    ) -> np.ndarray:
        """Compute the total linear_momentum of the system

        Parameters
        ----------
        sol_state : np.ndarray
            3D array of solution state for each snapshot, with shape
            (num_snapshots, num_particles, 7) being
            [m, x, y, z, vx, vy, vz]

        Returns
        -------
        linear_momentum : np.ndarray
            1D array of total linear_momentum for each snapshot
        """
        # Check the dimension and shape of sol_state
        if len(sol_state.shape) != 3:
            raise ValueError("sol_state must be a 3D array")

        if sol_state.shape[2] != 7:
            raise ValueError(
                "sol_state must have shape (num_snapshots, num_particles, 7)"
            )

        # Compute the total energy
        linear_momentum = np.zeros(sol_state.shape[0], dtype=np.float64)
        self.c_lib.compute_linear_momentum_python(
            linear_momentum.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
            sol_state.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
            ctypes.c_int32(sol_state.shape[0]),
            ctypes.c_int32(sol_state.shape[1]),
        )

        return linear_momentum

    def compute_angular_momentum(
        self,
        sol_state: np.ndarray,
    ) -> np.ndarray:
        """Compute the total angular_momentum of the system

        Parameters
        ----------
        sol_state : np.ndarray
            3D array of solution state for each snapshot, with shape
            (num_snapshots, num_particles, 7) being
            [m, x, y, z, vx, vy, vz]

        Returns
        -------
        angular_momentum : np.ndarray
            1D array of total angular_momentum for each snapshot
        """
        # Check the dimension and shape of sol_state
        if len(sol_state.shape) != 3:
            raise ValueError("sol_state must be a 3D array")

        if sol_state.shape[2] != 7:
            raise ValueError(
                "sol_state must have shape (num_snapshots, num_particles, 7)"
            )

        # Compute the total energy
        angular_momentum = np.zeros(sol_state.shape[0], dtype=np.float64)
        self.c_lib.compute_angular_momentum_python(
            angular_momentum.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
            sol_state.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
            ctypes.c_int32(sol_state.shape[0]),
            ctypes.c_int32(sol_state.shape[1]),
        )

        return angular_momentum

    @staticmethod
    def compute_eccentricity(
        G: float,
        sol_state: np.ndarray,
    ) -> np.ndarray:
        """Compute the eccentricity using the sol_state array,
        assuming that the first object is the central object

        Parameters
        ----------
        G : float
            Gravitational constant
        sol_state : np.ndarray
            Solution state of the system

        Returns
        -------
        np.ndarray
            Eccentricity of the system at each time step,
            with shape (num_snapshots, num_particles - 1)

        Notes
        -----
        - The function assumes that the first object is the central object.
        - C library function is not used here since this can be done with
            purely numpy vectorized operations. Nevertheless, we may consider
            implementing this in C library in the future.
        """
        num_snapshots = sol_state.shape[0]
        m_0 = sol_state[0, 0, 0]
        m = sol_state[0, 1:, 0]

        eccentricity = np.zeros(num_snapshots)

        x = sol_state[:, 1:, 1:4].copy() - sol_state[:, 0, 1:4].reshape(-1, 1, 3)
        v = sol_state[:, 1:, 4:7].copy() - sol_state[:, 0, 4:7].reshape(-1, 1, 3)

        denom = G * (m_0 + m)[np.newaxis, :, np.newaxis]
        eccentricity = (
            np.cross(v, np.cross(x, v)) / denom
            - x / np.linalg.norm(x, axis=2)[:, :, np.newaxis]
        )
        eccentricity = np.linalg.norm(eccentricity, axis=2)

        return eccentricity

    @staticmethod
    def compute_inclination(sol_state: np.ndarray) -> np.ndarray:
        """Compute the inclination using the sol_state array,
        assuming that the first object is the central object

        Parameters
        ----------
        sol_state : np.ndarray
            Solution state of the system

        Returns
        -------
        np.ndarray
            Inclination of the system at each time step,
            with shape (num_snapshots, num_particles - 1)

        Notes
        -----
        - The function assumes that the first object is the central object.
        - C library function is not used here since this can be done with
          purely numpy vectorized operations. Nevertheless, we may consider
          implementing this in C library in the future.
        """
        num_snapshots = sol_state.shape[0]

        inclination = np.zeros(num_snapshots)

        x = sol_state[:, 1:, 1:4].copy() - sol_state[:, 0, 1:4].reshape(-1, 1, 3)
        v = sol_state[:, 1:, 4:7].copy() - sol_state[:, 0, 4:7].reshape(-1, 1, 3)

        unit_angular_momentum_vector = (
            np.cross(x, v) / np.linalg.norm(np.cross(x, v), axis=2)[:, :, np.newaxis]
        )
        unit_z = np.array([0, 0, 1])

        inclination = np.arccos(np.sum(unit_angular_momentum_vector * unit_z, axis=2))

        return inclination

    @staticmethod
    def plot_rel_energy_error(
        sol_energy: np.ndarray,
        sol_time: np.ndarray,
        is_log_y: bool = True,
        title: Optional[str] = None,
        xlabel: Optional[str] = "Time",
        ylabel: Optional[str] = "$(E_0 - E(t)) / E_0$",
        save_fig: bool = False,
        save_fig_path: Optional[str | Path] = None,
    ) -> None:
        if sol_energy[0] == 0.0:
            warnings.warn("The initial energy is zero.")
        rel_energy_error = np.abs((sol_energy - sol_energy[0]) / sol_energy[0])
        plotting.plot_quantity_against_time(
            quantity=rel_energy_error,
            sol_time=sol_time,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            is_log_y=is_log_y,
            save_fig=save_fig,
            save_fig_path=save_fig_path,
        )

    @staticmethod
    def plot_rel_linear_momentum_error(
        sol_linear_momentum: np.ndarray,
        sol_time: np.ndarray,
        is_log_y: bool = True,
        title: Optional[str] = None,
        xlabel: Optional[str] = "Time",
        ylabel: Optional[str] = "Relative linear momentum error",
        save_fig: bool = False,
        save_fig_path: Optional[str | Path] = None,
    ) -> None:
        if sol_linear_momentum[0] == 0.0:
            warnings.warn("The initial linear momentum is zero.")
        rel_linear_momentum_error = np.abs(
            (sol_linear_momentum - sol_linear_momentum[0]) / sol_linear_momentum[0]
        )
        plotting.plot_quantity_against_time(
            quantity=rel_linear_momentum_error,
            sol_time=sol_time,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            is_log_y=is_log_y,
            save_fig=save_fig,
            save_fig_path=save_fig_path,
        )

    @staticmethod
    def plot_rel_angular_momentum_error(
        sol_angular_momentum: np.ndarray,
        sol_time: np.ndarray,
        is_log_y: bool = True,
        title: Optional[str] = None,
        xlabel: Optional[str] = "Time",
        ylabel: Optional[str] = "$(L_0 - L(t)) / L_0$",
        save_fig: bool = False,
        save_fig_path: Optional[str | Path] = None,
    ) -> None:
        if sol_angular_momentum[0] == 0.0:
            warnings.warn("The initial angular momentum is zero.")
        angular_momentum_error = np.abs(
            (sol_angular_momentum - sol_angular_momentum[0]) / sol_angular_momentum[0]
        )
        plotting.plot_quantity_against_time(
            quantity=angular_momentum_error,
            sol_time=sol_time,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            is_log_y=is_log_y,
            save_fig=save_fig,
            save_fig_path=save_fig_path,
        )



"""
Utility functions for grav_sim API
"""

import ctypes
from pathlib import Path
from typing import Optional

import numpy as np


def load_c_lib(c_lib_path: Optional[str | Path] = None) -> ctypes.CDLL:
    """Load the C dynamic-link library

    Returns
    -------
    c_lib : ctypes.CDLL
        C dynamic-link library object

    Raises
    ------
    OSError
        If the platform is not supported
    FileNotFoundError
        If the C library is not found at the path
    """
    if not c_lib_path:
        search_path = Path(__file__).parent.parent
        c_lib_files = [str(p) for p in search_path.rglob("*libgrav_sim*")]
        if len(c_lib_files) == 0:
            raise FileNotFoundError(f"C library not found from path: {search_path}")

        c_lib_path = c_lib_files[0]

    return ctypes.cdll.LoadLibrary(str(c_lib_path))


def initialize_c_lib(c_lib: ctypes.CDLL) -> None:
    """Initialize C library

    Parameters
    ----------
    c_lib : ctypes.CDLL
        C dynamic-link library object

    Raises
    ------
    ValueError
        If any C functions that should be available are not found in the C library
    """
    c_lib.launch_simulation_python.restype = int
    c_lib.launch_cosmological_simulation_python.restype = int
    c_lib.compute_energy_python.restype = None
    c_lib.compute_linear_momentum_python.restype = None
    c_lib.compute_angular_momentum_python.restype = None
    c_lib.load_built_in_system_python.restype = int
    c_lib.free_memory_int32.restype = None
    c_lib.free_memory_double.restype = None
    c_lib.keplerian_to_cartesian_python.restype = None


def keplerian_to_cartesian(
    c_lib: ctypes.CDLL,
    semi_major_axis: float,
    eccentricity: float,
    inclination: float,
    argument_of_periapsis: float,
    longitude_of_ascending_node: float,
    true_anomaly: float,
    total_mass: float,
    G: float,
) -> tuple[np.ndarray, np.ndarray]:
    """Convert Keplerian elements to Cartesian coordinates

    Parameters
    ----------
    semi_major_axis : float
        Semi-major axis
    eccentricity : float
        Eccentricity
    inclination : float
        Inclination
    raan : float
        Right ascension of ascending node
    arg_periapsis : float
        Argument of periapsis
    true_anomaly : float
        True anomaly
    total_mass : float
        Total mass of the two bodies
    G : float
        Gravitational constant

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Position and velocity vectors in Cartesian coordinates
    """

    x = ctypes.c_double()
    y = ctypes.c_double()
    z = ctypes.c_double()
    v_x = ctypes.c_double()
    v_y = ctypes.c_double()
    v_z = ctypes.c_double()
    c_lib.keplerian_to_cartesian_python(
        ctypes.byref(x),
        ctypes.byref(y),
        ctypes.byref(z),
        ctypes.byref(v_x),
        ctypes.byref(v_y),
        ctypes.byref(v_z),
        ctypes.c_double(semi_major_axis),
        ctypes.c_double(eccentricity),
        ctypes.c_double(inclination),
        ctypes.c_double(argument_of_periapsis),
        ctypes.c_double(longitude_of_ascending_node),
        ctypes.c_double(true_anomaly),
        ctypes.c_double(total_mass),
        ctypes.c_double(G),
    )

    return np.array([x.value, y.value, z.value]), np.array(
        [v_x.value, v_y.value, v_z.value]
    )



"""
Plotting module for the grav_sim package.
"""

from pathlib import Path
from typing import Optional

import numpy as np
import matplotlib.pyplot as plt

from . import utils

SOLAR_SYSTEM_COLORS = {
    "Sun": "orange",
    "Mercury": "slategrey",
    "Venus": "wheat",
    "Earth": "skyblue",
    "Mars": "red",
    "Jupiter": "darkgoldenrod",
    "Saturn": "gold",
    "Uranus": "paleturquoise",
    "Neptune": "blue",
    "Moon": "grey",
    "Pluto": None,
    "Ceres": None,
    "Vesta": None,
}


def set_3d_axes_equal(ax: plt.Axes) -> None:
    """
    Make axes of 3D plot have equal scale

    Parameters
    ----------
    ax : matplotlib axis
        The axis to set equal scale

    Reference
    ---------
    karlo, https://stackoverflow.com/questions/13685386/how-to-set-the-equal-aspect-ratio-for-all-axes-x-y-z
    """

    x_limits = ax.get_xlim3d()  # type: ignore
    y_limits = ax.get_ylim3d()  # type: ignore
    z_limits = ax.get_zlim3d()  # type: ignore

    x_range = abs(x_limits[1] - x_limits[0])
    x_middle = np.mean(x_limits)
    y_range = abs(y_limits[1] - y_limits[0])
    y_middle = np.mean(y_limits)
    z_range = abs(z_limits[1] - z_limits[0])
    z_middle = np.mean(z_limits)

    # The plot bounding box is a sphere in the sense of the infinity
    # norm, hence I call half the max range the plot radius.
    plot_radius = 0.5 * max([x_range, y_range, z_range])

    ax.set_xlim3d([x_middle - plot_radius, x_middle + plot_radius])  # type: ignore
    ax.set_ylim3d([y_middle - plot_radius, y_middle + plot_radius])  # type: ignore
    ax.set_zlim3d([z_middle - plot_radius, z_middle + plot_radius])  # type: ignore


def plot_2d_trajectory(
    sol_state: np.ndarray,
    colors: Optional[list[str]] = None,
    labels: Optional[list[str]] = None,
    legend: bool = False,
    xlabel: str = "$x$",
    ylabel: str = "$y$",
    title: Optional[str] = None,
    marker: str = "o",
    markersize: int = 6,
    save_fig: bool = False,
    save_fig_path: Optional[str | Path] = None,
) -> None:
    """Plot 2D trajectory of objects

    Parameters
    ----------
    sol_state : np.ndarray
        Solution state of the system
    colors : Optional[list[str]], optional
        Colors of the trajectories, by default None
    labels : Optional[list[str]], optional
        Labels of the objects, used for legend, by default None
    legend : bool, optional
        Flag to check whether to show legend, by default False
    xlabel : str, optional
        Label of x-axis, by default "$x$ (AU)"
    ylabel : str, optional
        Label of y-axis, by default "$y$ (AU)"
    marker : str, optional
        Marker for the last position, by default "o"
    markersize : int, optional
        Marker size for the last position, by default 6
    save_fig : bool, optional
        Flag to check whether to save the figure, by default False
    save_fig_path : Optional[str], optional
        Path to save the figure, by default None
    """
    fig = plt.figure()
    ax = fig.add_subplot(111, aspect="equal")

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    for i in range(sol_state.shape[1]):
        if colors is not None:
            colors_i = colors[i]
        else:
            colors_i = None

        if labels is not None:
            labels_i = labels[i]
        else:
            labels_i = None

        traj = ax.plot(
            sol_state[:, i, 1],
            sol_state[:, i, 2],
            color=colors_i,
        )
        # Plot the last position with marker
        ax.plot(
            sol_state[-1, i, 1],
            sol_state[-1, i, 2],
            color=traj[0].get_color(),
            label=labels_i,
            marker=marker,
            markersize=markersize,
        )

    if title is not None:
        ax.set_title(title)

    if legend:
        fig.legend(loc="center right", borderaxespad=0.2)
        fig.tight_layout()

    if save_fig:
        plt.savefig(save_fig_path, dpi=300)
    else:
        plt.show()
    plt.close("all")


def plot_3d_trajectory(
    sol_state: np.ndarray,
    colors: Optional[list[str]] = None,
    labels: Optional[list[str]] = None,
    legend: bool = False,
    xlabel: str = "$x$",
    ylabel: str = "$y$",
    zlabel: str = "$z$",
    title: Optional[str] = None,
    marker: str = "o",
    markersize: int = 6,
    save_fig: bool = False,
    save_fig_path: Optional[str | Path] = None,
) -> None:
    """Plot 3D trajectory of objects

    Parameters
    ----------
    sol_state : np.ndarray
        Solution state of the system
    colors : Optional[list[str]], optional
        Colors of the trajectories, by default None
    labels : Optional[list[str]], optional
        Labels of the objects, used for legend, by default None
    legend : bool, optional
        Flag to check whether to show legend, by default False
    xlabel : str, optional
        Label of x-axis, by default "$x$ (AU)"
    ylabel : str, optional
        Label of y-axis, by default "$y$ (AU)"
    zlabel : str, optional
        Label of z-axis, by default "$z$ (AU)"
    marker : str, optional
        Marker for the last position, by default "o"
    markersize : int, optional
        Marker size for the last position, by default 6
    save_fig : bool, optional
        Flag to check whether to save the figure, by default False
    save_fig_path : Optional[str], optional
        Path to save the figure, by default None
    """
    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")
    ax.set_box_aspect([1.0, 1.0, 1.0])  # type: ignore
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_zlabel(zlabel)  # type: ignore

    for i in range(sol_state.shape[1] // 6):
        if colors is not None:
            colors_i = colors[i]
        else:
            colors_i = None

        if labels is not None:
            labels_i = labels[i]
        else:
            labels_i = None

        traj = ax.plot(
            sol_state[:, i * 3],
            sol_state[:, i * 3 + 1],
            sol_state[:, i * 3 + 2],
            color=colors_i,
        )
        # Plot the last position with marker
        ax.plot(
            sol_state[-1, i * 3],
            sol_state[-1, i * 3 + 1],
            sol_state[-1, i * 3 + 2],
            color=traj[0].get_color(),
            label=labels_i,
            marker=marker,
            markersize=markersize,
        )

    set_3d_axes_equal(ax)

    if title is not None:
        ax.set_title(title)

    if legend:
        ax.legend(loc="center right", bbox_to_anchor=(1.325, 0.5))
        fig.subplots_adjust(
            right=0.7
        )  # Adjust the right boundary of the plot to make room for the legend

    fig.tight_layout()

    if save_fig:
        plt.savefig(save_fig_path, dpi=300)
    else:
        plt.show()
    plt.close("all")


def plot_quantity_against_time(
    quantity: np.ndarray,
    sol_time: np.ndarray,
    is_log_y: bool = False,
    title: Optional[str] = None,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    colors: Optional[list[str]] = None,
    labels: Optional[list[str]] = None,
    legend: bool = False,
    save_fig: bool = False,
    save_fig_path: Optional[str | Path] = None,
) -> None:
    """Plot a quantity against time

    Parameters
    ----------
    quantity : np.ndarray
        Quantity to be plotted
    sol_time : np.ndarray
        Solution time
    is_log_y : bool, optional
        Flag to check whether y-axis is in log scale, by default False
    title : Optional[str], optional
        Title of the plot, by default None
    xlabel : Optional[str], optional
        Label of x-axis, by default None
    ylabel : Optional[str], optional
        Label of y-axis, by default None
    colors : Optional[list[str]], optional
        Colors of the trajectories, by default None
    labels : Optional[list[str]], optional
        Labels of the objects, used for legend, by default None
    legend : bool, optional
        Flag to check whether to show legend, by default False
    save_fig : bool, optional
        Flag to check whether to save the figure, by default False
    save_fig_path : Optional[str], optional
        Path to save the figure, by default None
    """
    fig = plt.figure()
    ax = fig.add_subplot(111)

    if len(quantity.shape) == 1:
        quantity = quantity.reshape(-1, 1)

    for i in range(quantity.shape[1]):
        if colors is not None:
            colors_i = colors[i]
        else:
            colors_i = None

        if labels is not None:
            labels_i = labels[i]
        else:
            labels_i = None

        ax.plot(sol_time, quantity[:, i], color=colors_i, label=labels_i)

    if is_log_y:
        ax.set_yscale("log")
    if title is not None:
        ax.set_title(title)
    if xlabel is not None:
        ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)

    if legend:
        ax.legend(loc="center right", bbox_to_anchor=(1.325, 0.5))
        fig.subplots_adjust(
            right=0.7
        )  # Adjust the right boundary of the plot to make room for the legend

    if save_fig:
        plt.savefig(save_fig_path, dpi=300)
    else:
        plt.show()
    plt.close("all")



from pathlib import Path


class AccelerationParam:
    """Acceleration parameters for gravity simulation

    Attributes
    ----------
    method : str
        Acceleration method
    softening_length : float
        Softening length for acceleration
    pm_grid_size : int
        Grid size for Particle Mesh method
    opening_angle : float
        Opening angle for Barnes-Hut method
    max_num_particles_per_leaf : int
        Maximum number of particles per leaf for Barnes-Hut method
    """

    AVAILABLE_ACCELERATION_METHODS = [
        "pairwise",
        "massless",
        "barnes_hut",
        "particle_mesh",
    ]
    ACCELERATION_METHODS_TO_ENCODING = {
        "pairwise": 1,
        "massless": 2,
        "barnes_hut": 3,
        "particle_mesh": 4,
    }
    ACCELERATION_ENCODING_TO_METHODS = {
        v: k for k, v in ACCELERATION_METHODS_TO_ENCODING.items()
    }

    def __init__(self) -> None:
        self.method = "pairwise"
        self.opening_angle = 1.0
        self.softening_length = 0.0
        self.pm_grid_size = 128
        self.max_num_particles_per_leaf = 1

    @property
    def method(self) -> str:
        """Get the acceleration method"""
        return self.ACCELERATION_ENCODING_TO_METHODS[self._method]

    @method.setter
    def method(self, value: str) -> None:
        """Set the acceleration method"""
        value = value.lower().strip()
        if value not in self.AVAILABLE_ACCELERATION_METHODS:
            raise ValueError(
                f"Invalid acceleration method: {value}. Available methods: {self.AVAILABLE_ACCELERATION_METHODS}"
            )
        self._method: int = self.ACCELERATION_METHODS_TO_ENCODING[value]

    @property
    def opening_angle(self) -> float:
        """Get the opening angle for Barnes-Hut method"""
        return self._opening_angle

    @opening_angle.setter
    def opening_angle(self, value: float) -> None:
        """Set the opening angle for Barnes-Hut method"""
        value = float(value)
        if value <= 0:
            raise ValueError(f"Invalid opening angle: {value}. Must be positive.")
        self._opening_angle = value

    @property
    def softening_length(self) -> float:
        """Get the softening length"""
        return self._softening_length

    @softening_length.setter
    def softening_length(self, value: float) -> None:
        """Set the softening length"""
        value = float(value)
        if value < 0:
            raise ValueError(
                f"Invalid softening length: {value}. Must be non-negative."
            )
        self._softening_length = value

    @property
    def pm_grid_size(self) -> int:
        """Get the grid size for Particle Mesh method"""
        return self._pm_grid_size

    @pm_grid_size.setter
    def pm_grid_size(self, value: int) -> None:
        """Set the grid size for Particle Mesh method"""
        value = int(value)
        if value <= 0:
            raise ValueError(f"Invalid grid size: {value}. Must be positive.")
        self._pm_grid_size = value

    @property
    def max_num_particles_per_leaf(self) -> int:
        """Get the maximum number of particles per leaf for Barnes-Hut method"""
        return self._max_num_particles_per_leaf

    @max_num_particles_per_leaf.setter
    def max_num_particles_per_leaf(self, value: int) -> None:
        """Set the maximum number of particles per leaf for Barnes-Hut method"""
        value = int(value)
        if value <= 0:
            raise ValueError(
                f"Invalid maximum number of particles per leaf: {value}. Must be positive."
            )
        self._max_num_particles_per_leaf = value


class IntegratorParam:
    """Integrator parameters for gravity simulation

    Attributes
    ----------
    integrator : str
        Integrator
    dt : float
        Time step
    tolerance : float
        Tolerance for adaptive step size integrators
    initial_dt : float
        Initial time step for adaptive step size integrators
    whfast_remove_invalid_particles : bool
        Flag to indicate whether to remove invalid particles in WHFast integrator
    """

    AVAILABLE_INTEGRATORS = [
        "euler",
        "euler_cromer",
        "rk4",
        "leapfrog",
        "rkf45",
        "dopri",
        "dverk",
        "rkf78",
        "ias15",
        "whfast",
    ]
    FIXED_STEP_SIZE_INTEGRATORS = ["euler", "euler_cromer", "rk4", "leapfrog", "whfast"]
    ADAPTIVE_STEP_SIZE_INTEGRATORS = ["rkf45", "dopri", "dverk", "rkf78", "ias15"]

    INTEGRATOR_TO_ENCODING = {
        "euler": 1,
        "euler_cromer": 2,
        "rk4": 3,
        "leapfrog": 4,
        "rkf45": 5,
        "dopri": 6,
        "dverk": 7,
        "rkf78": 8,
        "ias15": 9,
        "whfast": 10,
    }
    ENCODING_TO_INTEGRATORS = {k: v for v, k in INTEGRATOR_TO_ENCODING.items()}

    # Recommended settings for built-in systems with IAS15 integrator
    RECOMMENDED_IAS15_SETTINGS_BUILT_IN_SYSTEMS = {
        # "template": ["tf", "tf unit", "tolerance", "storing_freq"],
        "circular_binary_orbit": [50.0, "days", 1e-9, 1],
        "eccentric_binary_orbit": [2.6, "years", 1e-9, 1],
        "3d_helix": [20.0, "days", 1e-9, 1],
        "sun_earth_moon": [1.0, "years", 1e-9, 1],
        "figure-8": [20.0, "days", 1e-9, 1],
        "pyth-3-body": [70.0, "days", 1e-9, 1],
        "solar_system": [200.0, "years", 1e-9, 1],
        "solar_system_plus": [250.0, "years", 1e-9, 1],
    }

    def __init__(self) -> None:
        self.integrator = "euler"
        self.dt = -1.0
        self.tolerance = -1.0
        self.initial_dt = -1.0
        self.whfast_remove_invalid_particles = True

    @property
    def integrator(self) -> str:
        """Get the integrator"""
        return self.ENCODING_TO_INTEGRATORS[self._integrator]

    @integrator.setter
    def integrator(self, value: str) -> None:
        """Set the integrator"""
        value = value.lower().strip()
        if value not in self.AVAILABLE_INTEGRATORS:
            raise ValueError(
                f"Invalid integrator: {value}. Available integrators: {self.AVAILABLE_INTEGRATORS}"
            )
        self._integrator: int = self.INTEGRATOR_TO_ENCODING[value]

    @property
    def dt(self) -> float:
        """Get the time step size"""
        return self._dt

    @dt.setter
    def dt(self, value: float) -> None:
        """Set the time step size"""
        self._dt = float(value)

    @property
    def tolerance(self) -> float:
        """Get the tolerance for adaptive step size integrators"""
        return self._tolerance

    @tolerance.setter
    def tolerance(self, value: float) -> None:
        """Set the tolerance for adaptive step size integrators"""
        self._tolerance = float(value)

    @property
    def initial_dt(self) -> float:
        """Get the initial time step size for adaptive step size integrators"""
        return self._initial_dt

    @initial_dt.setter
    def initial_dt(self, value: float) -> None:
        """Set the initial time step size for adaptive step size integrators"""
        self._initial_dt = float(value)

    @property
    def whfast_remove_invalid_particles(self) -> bool:
        """Get the flag for removing invalid particles in WHFast integrator"""
        return self._whfast_remove_invalid_particles

    @whfast_remove_invalid_particles.setter
    def whfast_remove_invalid_particles(self, value: bool) -> None:
        """Set the flag for removing invalid particles in WHFast integrator"""
        self._whfast_remove_invalid_particles = bool(value)


class OutputParam:
    """Output parameters for gravity simulation

    Attributes
    ----------
    method : str
        Output method
    output_dir : str
        Output directory
    output_initial : bool
        Flag to indicate whether to output initial state
    output_interval : float
        Output interval
    coordinate_output_dtype : str
        Data type for coordinate output
    velocity_output_dtype : str
        Data type for velocity output
    mass_output_dtype : str
        Data type for mass output
    """

    AVAILABLE_OUTPUT_METHODS = ["disabled", "csv", "hdf5"]
    AVAILABLE_OUTPUT_DTYPE = ["float", "double"]

    OUTPUT_METHODS_TO_ENCODING = {
        "disabled": 1,
        "csv": 2,
        "hdf5": 3,
    }
    OUTPUT_ENCODING_TO_METHODS = {v: k for k, v in OUTPUT_METHODS_TO_ENCODING.items()}

    OUTPUT_DTYPE_TO_ENCODING = {
        "float": 1,
        "double": 2,
    }
    OUTPUT_ENCODING_TO_DTYPE = {v: k for k, v in OUTPUT_DTYPE_TO_ENCODING.items()}

    def __init__(self) -> None:
        self.method = "disabled"
        self.output_dir = "tmp/"
        self.output_initial = True
        self.output_interval = 1.0
        self.coordinate_output_dtype = "double"
        self.velocity_output_dtype = "double"
        self.mass_output_dtype = "double"

    @property
    def method(self) -> str:
        """Get the output method"""
        return self.OUTPUT_ENCODING_TO_METHODS[self._method]

    @method.setter
    def method(self, value: str) -> None:
        """Set the output method"""
        value = value.lower().strip()
        if value not in self.AVAILABLE_OUTPUT_METHODS:
            raise ValueError(
                f"Invalid output method: {value}. Available methods: {self.AVAILABLE_OUTPUT_METHODS}"
            )
        self._method: int = self.OUTPUT_METHODS_TO_ENCODING[value]

    @property
    def output_dir(self) -> str:
        """Get the output directory"""
        return self._output_dir

    @output_dir.setter
    def output_dir(self, value: str) -> None:
        """Set the output directory"""
        value = str(Path(value).resolve())
        if not value.endswith("/"):
            value += "/"
        self._output_dir = value

    @property
    def output_initial(self) -> bool:
        """Get the flag for outputting initial state"""
        return self._output_initial

    @output_initial.setter
    def output_initial(self, value: bool) -> None:
        """Set the flag for outputting initial state"""
        self._output_initial = bool(value)

    @property
    def output_interval(self) -> float:
        """Get the output interval"""
        return self._output_interval

    @output_interval.setter
    def output_interval(self, value: float) -> None:
        """Set the output interval"""
        self._output_interval = float(value)

    @property
    def coordinate_output_dtype(self) -> str:
        """Get the data type for coordinate output"""
        return self.OUTPUT_ENCODING_TO_DTYPE[self._coordinate_output_dtype]

    @coordinate_output_dtype.setter
    def coordinate_output_dtype(self, value: str) -> None:
        """Set the data type for coordinate output"""
        value = value.lower().strip()
        if value not in self.AVAILABLE_OUTPUT_DTYPE:
            raise ValueError(
                f"Invalid coordinate output data type: {value}. Available types: {self.AVAILABLE_OUTPUT_DTYPE}"
            )
        self._coordinate_output_dtype: int = self.OUTPUT_DTYPE_TO_ENCODING[value]

    @property
    def velocity_output_dtype(self) -> str:
        """Get the data type for velocity output"""
        return self.OUTPUT_ENCODING_TO_DTYPE[self._velocity_output_dtype]

    @velocity_output_dtype.setter
    def velocity_output_dtype(self, value: str) -> None:
        """Set the data type for velocity output"""
        value = value.lower().strip()
        if value not in self.AVAILABLE_OUTPUT_DTYPE:
            raise ValueError(
                f"Invalid velocity output data type: {value}. Available types: {self.AVAILABLE_OUTPUT_DTYPE}"
            )
        self._velocity_output_dtype: int = self.OUTPUT_DTYPE_TO_ENCODING[value]

    @property
    def mass_output_dtype(self) -> str:
        """Get the data type for mass output"""
        return self.OUTPUT_ENCODING_TO_DTYPE[self._mass_output_dtype]

    @mass_output_dtype.setter
    def mass_output_dtype(self, value: str) -> None:
        """Set the data type for mass output"""
        value = value.lower().strip()
        if value not in self.AVAILABLE_OUTPUT_DTYPE:
            raise ValueError(
                f"Invalid mass output data type: {value}. Available types: {self.AVAILABLE_OUTPUT_DTYPE}"
            )
        self._mass_output_dtype: int = self.OUTPUT_DTYPE_TO_ENCODING[value]


class Settings:
    """Settings for gravity simulation

    Attributes
    ----------
    verbose : str
        Verbosity level
    enable_progress_bar : bool
        Flag to enable progress bar
    """

    AVAILABLE_VERBOSITY_LEVELS = ["ignore_all", "ignore_info", "normal", "verbose"]
    VERBOSITY_TO_ENCODING = {
        "ignore_all": 0,
        "ignore_info": 1,
        "normal": 2,
        "verbose": 3,
    }
    ENCODING_TO_VERBOSITY = {v: k for k, v in VERBOSITY_TO_ENCODING.items()}

    def __init__(self) -> None:
        self.verbose = "normal"
        self.enable_progress_bar = True

    @property
    def verbose(self) -> str:
        """Get the verbosity level"""
        return self.ENCODING_TO_VERBOSITY[self._verbose]

    @verbose.setter
    def verbose(self, value: str) -> None:
        """Set the verbosity level"""
        value = value.lower().strip()
        if value not in self.AVAILABLE_VERBOSITY_LEVELS:
            raise ValueError(
                f"Invalid verbosity level: {value}. Available levels: {self.AVAILABLE_VERBOSITY_LEVELS}"
            )
        self._verbose: int = self.VERBOSITY_TO_ENCODING[value]

    @property
    def enable_progress_bar(self) -> bool:
        """Get the flag for enabling progress bar"""
        return self._enable_progress_bar

    @enable_progress_bar.setter
    def enable_progress_bar(self, value: bool) -> None:
        """Set the flag for enabling progress bar"""
        self._enable_progress_bar = bool(value)
